/*
 * Copyright 2013 Canonical Ltd.
 *
 * This file is part of ubuntu-calculator-app.
 *
 * ubuntu-calculator-app is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3.
 *
 * ubuntu-calculator-app is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

Qt.include("engine.js");

var context = new Context;

function Formula() {
    var previous = new Token(0, T_NUMBER);
    var formula = '';
    var previousFormula = {};
    var previousFormulaCounter = -1;
    var lookupTbl = {
        '-': '−',
        '/': '÷',
        '*': '×'
    };
    var table = {
        "65" : "+",
        "66" : "-",
        "67" : "*",
        "68" : "/"
    }

    var isLastResult = false;   // This var becames true if the last button pressed by user is equal
    var isChangeSign = false;   // This var becames true if user want to change the sign of a result

    function operatorToString(digit) {
        if(lookupTbl[digit] !== undefined) {
            return lookupTbl[digit];
        } else {
            return digit;
        }
    }

    // Check if the given digit is one of the valid operators or not.
    function isOperator(digit){
        switch(digit){
        case "+":
        case "-":
        case "*":
        case "/":
        case "%":
        case "^":
            return true;
        default:
            return false;
        }
    }

    var _calculate = function(func){
        try{
            var result = parse(func, context);
            if (result === Number.POSITIVE_INFINITY)
                result = '∞';
            else if (result === Number.NEGATIVE_INFINITY)
                result = '-∞';
        }catch(exception){
            if(exception instanceof DivisionByZeroError){
                result = "division by zero error";
            } else{
                throw new SyntaxError("malformed calculation");
            }
        }
        return result;
    }

    this.push = function(digit){
        var last = (new Scanner(formula + digit, context)).tokens.pop();

        // After a calc you have to add a sign before add a number
        if (!isOperator(digit) && isLastResult) return;

        switch(last.type){
        case T_PLUS:
        case T_MINUS:
        case T_DIV:
        case T_TIMES:
            if (previous.type & T_OPERATOR && !(previous.type !== '67' && hasToAddDot)) return 'invalidOperation'; // Not two operator below
            break;
        case T_UNARY_MINUS:
            // An unary minus can not be after an unary minus or an unary plus
            if (previous.type === T_UNARY_MINUS || previous.type === T_UNARY_PLUS) return 'invalidOperation';
            if (previous.type === T_MINUS && formula.length === 1) {
                // If formula.lenght is 1 it's the first minus, so we have to delete it. In all other cases we add it
                formulaPop();
                return 'invalidOperation';
            }

            break;
        case T_UNARY_PLUS:
            // An unary plus can not be after a minus or a plus or a * or a /
            if (previous.type === T_UNARY_MINUS
                    || previous.type === T_UNARY_PLUS
                    || previous.type === T_PLUS
                    || previous.type === T_TIMES
                    || previous.type === T_DIV
                    || previous.type === T_MINUS
                ) return 'invalidOperation';
            break;
        }

        formulaView.headerItem.state = "newPush"; // This is to change "AC" to "C" when user continue a calc

        formula += digit;
        isLastResult = false;
        previous = last;
        var ret = {value: '', type: last.type};
        //formula Shouldn't start with operators.
        var firstChar = formula.substring(0,1);
        if(isNaN(firstChar) && firstChar !== "(" && firstChar !== "-"){ // First letter should be digit or open bracket or minus
            ret.type = T_OPERATOR // Return
            formula = ""; // Make formula var empty
        }
        ret.value = operatorToString(digit);
        return ret;
    };

    this.pop = function(){
        var scannedFormula = new Scanner(formula, context);
        var last = {type:T_NUMBER};
        var endRemoveSize = 0;
        var beginRemoveSize = 0;
        do {
            last = scannedFormula.tokens.pop();
            endRemoveSize += last.value.toString().length;
            //If we found close bracket, then we must look for opening bracket
            if (last.type === T_PCLOSE) {
                //Let's check if we have opening bracket at beginning
                if (formula[0] === '(') {
                    beginRemoveSize = 1;
                } else {
                    console.log("Error: Expecting that first character in formula will be '(' in formula: " +formula);
                }
            }
        } while (scannedFormula.tokens.length > 0 &&
                (last.type === T_NUMBER || last.type === T_UNARY_PLUS || last.type === T_UNARY_MINUS || last.type === T_PCLOSE));
        formula = formula.substring(beginRemoveSize, formula.length - endRemoveSize);
        previous = scannedFormula.tokens.pop();
    };

    this.numeralPop = function() {
        var scannedFormula = new Scanner(formula, context);
        var last = {type:T_NUMBER};
        var endRemoveSize = 0;
        var beginRemoveSize = 0;
        do {
            last = scannedFormula.tokens.pop();
            //If the token is the number, then delete only last numeral
            if (last.type === T_NUMBER) {
                endRemoveSize += 1;
            } else {
                endRemoveSize += last.value.toString().length;
            }
            //If we found close bracket, then we must look for opening bracket
            if (last.type === T_PCLOSE) {
                //Let's check if we have opening bracket at beginning
                if (formula[0] === '(') {
                    beginRemoveSize = 1;
                } else {
                    console.log("Error: Expecting that first character in formula will be '(' in formula: " +formula);
                }
            }
        } while (scannedFormula.tokens.length > 0 &&
                (last.type === T_UNARY_PLUS || last.type === T_UNARY_MINUS || last.type === T_PCLOSE));
        formula = formula.substring(beginRemoveSize, formula.length - endRemoveSize);
        previous = scannedFormula.tokens.pop();
    };


    this.calculate = function() {
        var result;
        var previousSize = previous.value.toString().length;
        if(formula.length <= previousSize) return;
        if (isLastResult) {
            var scannedFormula = new Scanner(previousFormula[previousFormulaCounter], context);
        } else {
            var scannedFormula = new Scanner(formula, context);
        }

        // Last and secondLast are one a number and the other an operator. Both can be both, it depends on user input
        var last = scannedFormula.tokens.pop();
        var secondLast = scannedFormula.tokens.pop();

        // In this case user press equal twice E.G. 2*2=4=8
        if (isLastResult && !isChangeSign && table[secondLast.type] !== undefined) {
            screenFormula.push({_text:'', _operation: operatorToString(table[secondLast.type]), _number: last.value.toString()});
            formula = formula + table[secondLast.type] + last.value.toString();
        }
        // In this case user press a number, and operator (or only an operator, if it isn't the first calc), and then equal
        else if (last.type & T_OPERATOR) {
            //if user has not put any number after last sign then ignore it
            formulaPop()
        }
        result = _calculate(formula);

        // If user want to change sign to the result
        if (isLastResult && isChangeSign) {
            result = -result;   // Change the sign
        }

        // Reset formula and add previous result to the array
        previousFormulaCounter++;
        previousFormula[previousFormulaCounter] = formula;
        formula = result;

        isLastResult = true;
        return result;
    }

    this.changeSign = function() {
        switch(previous.type){
        case T_PLUS:
        case T_MINUS:
        case T_DIV:
        case T_TIMES:
            // After an operator, push minus
            formulaPush("-");
            break;
        default:
            var scannedFormula = new Scanner(formula, context);
            var last = {type: undefined};
            var toAdd = [];
            var addMinus = true;

            // If is the first pressed button
            if ((scannedFormula.tokens.length === 0) && (previousFormulaCounter === -1)) {                   
                formulaPush('-');
                return;
            }

            // Create a queue with all element to re-add to var formula
            while(scannedFormula.tokens.length > 0 && last.type !== T_PLUS &&
                  last.type !== T_MINUS &&
                  last.type !== T_DIV &&
                  last.type !== T_TIMES ){
                last = scannedFormula.tokens.pop();
                if(last.type !== T_UNARY_MINUS && last.type !== T_UNARY_PLUS) {
                    toAdd.push(last);
                } else {
                    if(last.type === T_UNARY_MINUS)
                        addMinus = false;
                }
            }

            formulaPop();

            // This is for the first number only
            if (toAdd.length === 1) {
                var number = toAdd.pop();
                if (!isNaN(number.value)) { // Check if is a number
                    if (addMinus)
                        formulaPush('-');   // We add a minus
                    formulaPush(number.value.toString());

                    if (hasToAddDot)
                        formulaPush('.');

                    return; // We have finished
                }
                toAdd.push(number); // If number is not a number, we push it in queue for next function
            }


            while(toAdd.length > 0){
                var el = toAdd.pop()
                formulaPush(el.value.toString());
                if((el.type & T_OPERATOR) && addMinus){
                    formulaPush('-');
                }
            }
            if (hasToAddDot)
                formulaPush('.');
        }
    }

    this.hasResults = function(){
        if(results.length > 0){
            return true;
        }
        return false;
    }

    this.restorePreviousFormula = function() {
        if (previousFormulaCounter > -1) {
            formula = previousFormula[previousFormulaCounter];
            previousFormulaCounter--;
        }
    }
}
