# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2013 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Calculator app autopilot emulators."""

import logging

import autopilot.logging
from autopilot.introspection import dbus

from ubuntuuitoolkit import emulators as toolkit_emulators


logger = logging.getLogger(__name__)


class MainView(toolkit_emulators.MainView):
    """Calculator MainView Autopilot emulator."""

    # TODO when we start the tests for the scientific calculator, many of this
    # methods will have to be moved to the SimplePage emulator, and the main
    # view can call them depending on the current page displayed.
    # --elopio - 2013-08-08

    def __init__(self, *args):
        super(MainView, self).__init__(*args)
        self.pointing_device = toolkit_emulators.get_pointing_device()

    def clear_with_button(self):
        """Clear the calculation using the clear button."""
        self.get_calc_keyboard().click_clear()

    def get_calc_keyboard(self):
        """Return the CalcKeyboard autopilot emulator."""
        return self.select_single(CalcKeyboard)

    def calculate_operation(self, operation):
        """Calculate an operation.

        :parameter operation: A string with the operation to calculate.

        """
        self.enter_operation(operation)
        self.get_calc_keyboard().click_equals()

    def enter_operation(self, operation):
        """Enter an operation.

        :parameter operation: A string with the operation to enter.

        """
        self.get_calc_keyboard().enter_operation(operation)

    def change_sign(self):
        """Change the sign of the last operand."""
        self.get_calc_keyboard().click_sign()

    def get_result(self):
        """Return the result of a calculation."""
        return self.get_current_screen().get_result()

    def get_current_screen(self):
        """Return the screen for the current calculation."""
        count = self.get_number_of_screens()
        return self.select_many(Screen)[count - 1]

    def get_operand_by_index(self, index):
        """Return an operand of a calculation entered."""
        return self.get_current_screen().get_operand_by_index(index)

    def clear_with_swipe(self):
        """Clear the calculation making a swipe on the screen."""
        self.get_current_screen().swipe_up()
        self._wait_for_screen_to_finish_animation()

    def _wait_for_screen_to_finish_animation(self):
        qquicklistview = self.select_single('QQuickListView')
        qquicklistview.moving.wait_for(False)

    def is_cleared(self):
        """Return True if the screen has no calculation. False otherwise."""
        return self.get_current_screen().is_cleared()

    def get_screen_by_index(self, index):
        """Return a screen."""
        return self._get_screens()[index]

    def _get_screens(self):
        return self.select_many(Screen)

    def get_number_of_screens(self):
        """Return the number of screens."""
        return len(self._get_screens())

    def swipe_previous_calculation_into_view(self, screen_index):
        previous_screen = self.get_screen_by_index(screen_index)
        _, prev_screen_y, _, prev_screen_h = previous_screen.globalRect
        x, y, w, h = self.globalRect
        start_x = stop_x = x + w / 2
        start_y = y + h / 2
        stop_y = start_y + y - prev_screen_y
        self.pointing_device.drag(start_x, start_y, stop_x, stop_y)
        self._wait_for_screen_to_finish_animation()

    def delete_previous_calculation(self, screen_index, confirm):
        """Delete an old calculation.

        :parameter screen_index: If index of the screen with the calculation
            to delete.
        :parameter confirm: If True, the deletion will be confimed. If False,
            it will be canceled.

        """
        self.swipe_previous_calculation_into_view(screen_index)
        screen = self.get_screen_by_index(screen_index)
        screen.swipe_to_delete()
        self._wait_for_screen_to_finish_animation()
        if confirm:
            screen.confirm_removal()
        else:
            raise NotImplementedError('Not yet implemented.')


class CalcKeyboard(toolkit_emulators.UbuntuUIToolkitEmulatorBase):
    """CalcKeyboard Autopilot emulator."""

    _BUTTONS_DICT = {
        "0": "zeroButton",
        "1": "oneButton",
        "2": "twoButton",
        "3": "threeButton",
        "4": "fourButton",
        "5": "fiveButton",
        "6": "sixButton",
        "7": "sevenButton",
        "8": "eightButton",
        "9": "nineButton",
        "+": "plusButton",
        "-": "minusButton",
        "/": "divideButton",
        "*": "multiplyButton",
        '.': "pointButton",
    }

    def __init__(self, *args):
        super(CalcKeyboard, self).__init__(*args)
        self.pointing_device = toolkit_emulators.get_pointing_device()

    def click_clear(self):
        """Click the clear button."""
        clear_button = self._get_keyboard_button('clearButton')
        self.pointing_device.click_object(clear_button)

    def _get_keyboard_button(self, objectName):
        return self.select_single('KeyboardButton', objectName=objectName)

    def enter_operation(self, operation):
        """Enter an operation on the calculator.

        :parameter operation: A string with the operation to enter.

        """
        for char in operation:
            button = self._get_keyboard_button(self._BUTTONS_DICT[char])
            self.pointing_device.click_object(button)

    def click_equals(self):
        """Click the equals button."""
        equals_button = self._get_keyboard_button('equalsButton')
        self.pointing_device.click_object(equals_button)

    def click_sign(self):
        """Click the sign button."""
        sign_button = self._get_keyboard_button('signButton')
        self.pointing_device.click_object(sign_button)


class Screen(toolkit_emulators.Standard):
    """Screen Autopilot emulator."""

    def __init__(self, *args):
        super(Screen, self).__init__(*args)
        self.pointing_device = toolkit_emulators.get_pointing_device()

    def get_result(self):
        """Return the result of a calculation."""
        result_label = self.get_result_label()
        assert result_label is not None, 'No result displayed.'
        return result_label.numbers

    def get_result_label(self):
        try:
            return self.select_single(
                'CalcLabel', objectName='result', isLast=True)
        except dbus.StateNotFoundError:
            results = self.select_many('CalcLabel', objectName='result')
            count = len(results)
            if count > 0:
                return results[count - 1]
            else:
                return None

    def click_result_label(self):
        result = self.get_result_label()
        self.pointing_device.click_object(result)

    def get_edit_icon(self):
        results = self.select_many('QQuickItem', objectName='editIcon')
        count = len(results)
        if count > 0:
            return results[count - 1]
        else:
            return None

    def click_edit_icon(self):
        result = self.get_edit_icon()
        self.pointing_device.click_object(result)

    def get_title_label(self):
        return self.select_single('TextField', objectName='title')

    def click_title_label(self):
        text_label = self.get_title_label()
        self.pointing_device.click_object(text_label)

    def get_operand_by_index(self, index):
        """Return an operand of a calculation entered."""
        operand_label = self.select_single(
            'CalcLabel', objectName='operand{0}'.format(index))
        assert operand_label is not None, 'No operand with index {0}.'.format(
            index)
        return operand_label.numbers

    def swipe_up(self):
        """Swipe up on the screen to tear off the calculation."""
        x, y, w, h = self.globalRect
        tx = x + (w / 2)
        ty = y + (h / 2)

        self.pointing_device.drag(tx, ty + h, tx, ty - h)

    def is_cleared(self):
        """Return True if the screen has no calculation. False otherwise."""
        return (self.get_result_label() is None and
                self._get_number_of_labels() == 1 and
                self.get_operand_by_index(0) == "")

    def _get_number_of_labels(self):
        return len(self.select_many('CalcLabel'))

    @autopilot.logging.log_action(logger.info)
    def swipe_to_delete(self):
        """Swipe the screen to delete the calculation."""
        # We override it because the left side of the screen is used to
        # add labels, so the swipe has to start where the operation starts.
        # This is hard for a user to do, and weird for automation, which means
        # it's an usability issue. See http://pad.lv/1329536
        self._drag_pointing_device_to_delete()
        self.waitingConfirmationForRemoval.wait_for(True)

    def _drag_pointing_device_to_delete(self):
        x, y, width, height = self.globalRect
        left_x = x + (width * 0.33)
        right_x = x + (width * 0.9)
        start_y = stop_y = y + (height // 2)

        start_x = left_x
        stop_x = right_x
        self.pointing_device.drag(start_x, start_y, stop_x, stop_y)
