# -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
#
# Copyright (C) 2013 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Tests for the Calculator App"""

from __future__ import absolute_import

from testtools.matchers import Equals
from autopilot.matchers import Eventually
from autopilot.platform import model

from ubuntu_calculator_app.tests import CalculatorTestCase
import subprocess


class TestSimplePage(CalculatorTestCase):

    def setUp(self):
        super(TestSimplePage, self).setUp()

    @classmethod
    def setUpClass(cls):
        if model() != "Desktop":
            try:
                subprocess.check_call(["stop", "-q", "maliit-server"])
            except subprocess.CalledProcessError:
                pass

    @classmethod
    def tearDownClass(cls):
        if model() != "Desktop":
            try:
                subprocess.check_call(["start", "-q", "maliit-server"])
            except subprocess.CalledProcessError:
                pass

    def test_operation_after_clear(self):
        """ Test that after clearing one calculation, the next is still correct
        (bug #1164973).

        """
        self.main_view.calculate_operation("1+1")
        self._assert_result("2")
        self.main_view.clear_with_button()
        self.main_view.calculate_operation("1+1")
        self._assert_result("2")

    def _assert_result(self, expected_result):
        self.assertThat(
            self.main_view.get_result, Eventually(Equals(expected_result)))

    def test_enter_operand(self):
        self.main_view.enter_operation("123")
        self.assertEqual(self.main_view.get_operand_by_index(0), "123")

    def test_equals_dont_change_numbers(self):
        """ Test that typing a number and pressing "=" won't change the number
        (bug #1165894).

        """
        self.main_view.enter_operation("123")
        self.main_view.get_calc_keyboard().click_equals()

        # check that the label is still displaying correctly
        self.assertEqual(self.main_view.get_operand_by_index(0), "123")

    def test_multiply_priority(self):
        """Multiply is done before addition or substraction."""
        self.main_view.calculate_operation("2+2*2")
        self._assert_result("6")

    def test_divide_priority(self):
        """Divide is done before addition or substraction."""
        self.main_view.calculate_operation("5+6/2")
        self._assert_result("8")

    def test_divide_with_infinity_length_result_number(self):
        """Check how is formatted infinity long result number."""
        self.main_view.calculate_operation("1/3")
        self._assert_result("0.33333333")

    def test_swipe_up_clears(self):
        """Make sure swiping up clears the formula label."""
        self.main_view.enter_operation("9")
        self.main_view.clear_with_swipe()
        self.assertThat(self.main_view.is_cleared, Eventually(Equals(True)))

    def test_sign_button(self):
        """Ensure the sign button is working."""
        calc_keyboard = self.main_view.get_calc_keyboard()
        calc_keyboard.enter_operation("5+")
        calc_keyboard.click_sign()
        calc_keyboard.enter_operation("1")
        calc_keyboard.click_equals()

        self._assert_result("4")

    def test_addition(self):
        """Addition test."""
        self.main_view.calculate_operation("4+2.55")
        self._assert_result("6.55")

    def test_substraction(self):
        self.main_view.calculate_operation("5-4")
        self._assert_result("1")

    def test_positive_numbers_multiplication(self):
        self.main_view.calculate_operation("2.1*3")
        self._assert_result("6.3")

    def test_multiplication_by_zero(self):
        self.main_view.calculate_operation("4*0")
        self._assert_result("0")

    def test_one_negative_number_multiplication(self):
        # tested operation: -3*4
        calc_keyboard = self.main_view.get_calc_keyboard()
        calc_keyboard.click_sign()  # to get "-"
        calc_keyboard.enter_operation("3*4")
        calc_keyboard.click_equals()
        self._assert_result("-12")

    def test_two_negative_numbers_multiplication(self):
        # tested operation: -2*-3
        calc_keyboard = self.main_view.get_calc_keyboard()
        calc_keyboard.click_sign()  # to get "-"
        calc_keyboard.enter_operation("2*")
        calc_keyboard.click_sign()  # to get "-"
        calc_keyboard.enter_operation("3")
        calc_keyboard.click_equals()

        self._assert_result("6")

    def test_three_positive_numbers_multiplication(self):
        self.main_view.calculate_operation("2*3*10")
        self._assert_result("60")

    def test_two_negatives_in_three_numbers_multiplication(self):
        # tested operation: -2*-3*5
        calc_keyboard = self.main_view.get_calc_keyboard()
        calc_keyboard.click_sign()  # to get "-"
        calc_keyboard.enter_operation("2*")
        calc_keyboard.click_sign()  # to get "-"
        calc_keyboard.enter_operation("3*5")
        calc_keyboard.click_equals()

        self._assert_result("30")

    def test_three_negative_numbers_multiplication(self):
        # tested operation: -4*-5*-7
        calc_keyboard = self.main_view.get_calc_keyboard()
        calc_keyboard.click_sign()  # to get "-"
        calc_keyboard.enter_operation("4*")
        calc_keyboard.click_sign()  # to get "-"
        calc_keyboard.enter_operation("5*")
        calc_keyboard.click_sign()  # to get "-"
        calc_keyboard.enter_operation("7")
        calc_keyboard.click_equals()

        self._assert_result("-140")

    def test_divide_positive(self):
        self.main_view.calculate_operation("4/2")
        self._assert_result("2")

    def test_divide_negative(self):
        # tested operation: 4/-2
        calc_keyboard = self.main_view.get_calc_keyboard()
        calc_keyboard.enter_operation("4/")
        calc_keyboard.click_sign()  # to get "-"
        calc_keyboard.enter_operation("2")
        calc_keyboard.click_equals()

        self._assert_result("-2")

    def test_divide_two_negatives(self):
        # tested operation: -4/-2
        calc_keyboard = self.main_view.get_calc_keyboard()
        calc_keyboard.click_sign()  # to get "-"
        calc_keyboard.enter_operation("4/")
        calc_keyboard.click_sign()  # to get "-"
        calc_keyboard.enter_operation("2")
        calc_keyboard.click_equals()

        self._assert_result("2")

    def test_divide_with_zero(self):
        self.main_view.calculate_operation("0/4")
        self._assert_result("0")

    def test_divide_by_zero(self):
        self.main_view.calculate_operation("4/0")
        self._assert_result(u'+\u221e')

    def test_divide_zero_by_zero(self):
        self.main_view.calculate_operation("0/0")
        self._assert_result("NaN")

    def test_adding_small_number(self):
        self.main_view.calculate_operation("0.000000001+1")
        self._assert_result("1.000000001")

    def test_divide_small_number(self):
        """Check result number formating."""
        self.main_view.calculate_operation(".000000001/10")
        self._assert_result("1e-10")

    def test_large_numbers_multiplication(self):
        self.main_view.calculate_operation("99999999999*99999999999")
        self._assert_result("1.0000000e+22")

    def test_swipe_up_with_result(self):
        """Test swipe up adds calculation to history"""
        self.main_view.calculate_operation("9+8")
        self.main_view.clear_with_swipe()

        self.assertThat(
            self.main_view.get_number_of_screens, Eventually(Equals(2)))
        self.assertTrue(
            self.main_view.get_screen_by_index(1).is_cleared())
        self.assertEquals(
            self.main_view.get_screen_by_index(0).get_result(), "17")

    def test_swipe_to_delete_calculation(self):
        """Delete a calculation swiping on it to the side."""
        self.main_view.calculate_operation("9+5")
        self.main_view.clear_with_swipe()

        self.main_view.swipe_previous_calculation_into_view(0)
        self.main_view.delete_previous_calculation(0, confirm=True)

        self.assertThat(
            self.main_view.get_number_of_screens, Eventually(Equals(1)))

    def test_operands_before_number(self):
        """ Operands before the numbers."""
        self.main_view.calculate_operation("*3")
        screen = self.main_view.get_current_screen()
        result_label = screen.get_result_label()

        self.assertThat(result_label, Equals(None))

    def test_continue_calculation(self):
        self.main_view.calculate_operation("59+1")

        calc_keyboard = self.main_view.get_calc_keyboard()
        calc_keyboard.enter_operation("+5")
        calc_keyboard.click_equals()

        self._assert_result("65")

    def test_unfocus_label(self):
        """ Clear label focus by clicking outside its area. """
        self.main_view.calculate_operation("12+3")
        self._assert_result("15")

        # Focus the item
        screen = self.main_view.get_current_screen()
        screen.click_title_label()
        title_label = screen.get_title_label()
        self.assertThat(title_label.activeFocus, Eventually(Equals(True)))

        # Click elsewhere to unfocus the label
        screen.click_result_label()
        self.assertThat(title_label.activeFocus, Eventually(Equals(False)))

    def test_hide_calc_keyboard(self):
        """ Hide calc keyboard when focus on a label """
        self.main_view.calculate_operation("18-1")
        self._assert_result("17")

        # Focus the item
        screen = self.main_view.get_current_screen()
        screen.click_title_label()

        calc_keyboard = self.main_view.get_calc_keyboard()
        self.assertThat(calc_keyboard.visible, Eventually(Equals(False)))

        # Click elsewhere to unfocus the label
        screen.click_result_label()
        self.assertThat(calc_keyboard.visible, Eventually(Equals(True)))

    def test_edit_mode(self):
        """ Enter and exit edit mode clicking on pencil icon """
        self.main_view.calculate_operation("105-5")
        self._assert_result("100")

        # Click the edit button
        screen = self.main_view.get_current_screen()
        screen.click_edit_icon()

        # Check if is focused
        title_label = screen.get_title_label()
        self.assertThat(title_label.activeFocus, Eventually(Equals(True)))

        # Click to unfocus the label
        screen.click_edit_icon()
        self.assertThat(title_label.activeFocus, Eventually(Equals(False)))
